@@ -1,21 +1,26 @@
 from __future__ import unicode_literals
 
-from .util import differences
-from .statements import Statements
-from functools import partial
 from collections import OrderedDict as od
+from functools import partial
+
+from .statements import Statements
+from .util import differences
 
 THINGS = [
-    'schemas',
-    'enums',
-    'sequences',
-    'constraints',
-    'functions',
-    'views',
-    'indexes',
-    'extensions',
+    "schemas",
+    "enums",
+    "sequences",
+    "constraints",
+    "functions",
+    "views",
+    "indexes",
+    "extensions",
+    "privileges",
+    "collations",
+    "rlspolicies",
+    "triggers",
 ]
-PK = 'PRIMARY KEY'
+PK = "PRIMARY KEY"
 
 
 def statements_for_changes(
@@ -28,17 +33,37 @@ def statements_for_changes(
     add_dependents_for_modified=False,
 ):
     added, removed, modified, unmodified = differences(things_from, things_target)
-    if add_dependents_for_modified:
-        for k, m in list(modified.items()):
-            for d in m.dependents_all:
-                if d in unmodified:
-                    modified[d] = unmodified.pop(d)
-        modified = od(sorted(modified.items()))
+
+    return statements_from_differences(
+        added=added,
+        removed=removed,
+        modified=modified,
+        replaceable=None,
+        creations_only=creations_only,
+        drops_only=drops_only,
+        modifications=modifications,
+        dependency_ordering=dependency_ordering,
+        old=things_from,
+    )
+
+
+def statements_from_differences(
+    added,
+    removed,
+    modified,
+    replaceable=None,
+    creations_only=False,
+    drops_only=False,
+    modifications=True,
+    dependency_ordering=False,
+    old=None,
+):
+    replaceable = replaceable or set()
     statements = Statements()
     if not creations_only:
         pending_drops = set(removed)
         if modifications:
-            pending_drops |= set(modified)
+            pending_drops |= set(modified) - replaceable
     else:
         pending_drops = set()
     if not drops_only:
@@ -66,7 +91,7 @@ def statements_for_changes(
             for k, v in removed.items():
                 if not has_remaining_dependents(v, pending_drops):
                     if k in pending_drops:
-                        statements.append(v.drop_statement)
+                        statements.append(old[k].drop_statement)
                         pending_drops.remove(k)
         if not drops_only:
             for k, v in added.items():
@@ -79,7 +104,7 @@ def statements_for_changes(
                 if not creations_only:
                     if not has_remaining_dependents(v, pending_drops):
                         if k in pending_drops:
-                            statements.append(v.drop_statement)
+                            statements.append(old[k].drop_statement)
                             pending_drops.remove(k)
                 if not drops_only:
                     if not has_uncreated_dependencies(v, pending_creations):
@@ -90,8 +115,10 @@ def statements_for_changes(
         if not after:
             break
 
-        elif after == before:  # this should never happen because there shouldn't be circular dependencies
-            raise ValueError('cannot resolve dependencies')  # pragma: no cover
+        elif (
+            after == before
+        ):  # this should never happen because there shouldn't be circular dependencies
+            raise ValueError("cannot resolve dependencies")  # pragma: no cover
 
     return statements
 
@@ -108,7 +135,11 @@ def get_enum_modifications(tables_from, tables_target, enums_from, enums_target)
         _, _, c_modified, _ = differences(t_before.columns, v.columns)
         for k, c in c_modified.items():
             before = t_before.columns[k]
-            if c.is_enum == before.is_enum and c.dbtypestr == before.dbtypestr and c.enum != before.enum:
+            if (
+                c.is_enum == before.is_enum
+                and c.dbtypestr == before.dbtypestr
+                and c.enum != before.enum
+            ):
                 pre.append(before.change_enum_to_string_statement(t))
                 post.append(before.change_string_to_enum_statement(t))
     for e in enums_to_change.values():
@@ -117,8 +148,9 @@ def get_enum_modifications(tables_from, tables_target, enums_from, enums_target)
     return pre + recreate + post
 
 
-def get_schema_changes(tables_from, tables_target, enums_from, enums_target):
+def get_table_changes(tables_from, tables_target, enums_from, enums_target):
     added, removed, modified, _ = differences(tables_from, tables_target)
+
     statements = Statements()
     for t, v in removed.items():
         statements.append(v.drop_statement)
@@ -127,9 +159,39 @@ def get_schema_changes(tables_from, tables_target, enums_from, enums_target):
     statements += get_enum_modifications(
         tables_from, tables_target, enums_from, enums_target
     )
+
     for t, v in modified.items():
         before = tables_from[t]
+
+        # drop/recreate tables which have changed from partitioned to non-partitioned
+        if v.is_partitioned != before.is_partitioned:
+            statements.append(v.drop_statement)
+            statements.append(v.create_statement)
+            continue
+
+        # attach/detach tables with changed parent tables
+        if v.parent_table != before.parent_table:
+            statements += v.attach_detach_statements(before)
+
+    for t, v in modified.items():
+        before = tables_from[t]
+
+        if not v.is_alterable:
+            continue
+
         c_added, c_removed, c_modified, _ = differences(before.columns, v.columns)
+
+        for k in list(c_modified):
+            c = v.columns[k]
+            c_before = before.columns[k]
+
+            # there's no way to alter a table into/out of generated state
+            # so you gotta drop/recreate
+            if c.is_generated != c_before.is_generated:
+                del c_modified[k]
+                c_added[k] = c
+                c_removed[k] = c_before
+
         for k, c in c_removed.items():
             alter = v.alter_table_statement(c.drop_column_clause)
             statements.append(alter)
@@ -138,50 +200,121 @@ def get_schema_changes(tables_from, tables_target, enums_from, enums_target):
             statements.append(alter)
         for k, c in c_modified.items():
             statements += c.alter_table_statements(before.columns[k], t)
+
+        if v.rowsecurity != before.rowsecurity:
+            rls_alter = v.alter_rls_statement
+            statements += [rls_alter]
+    return statements
+
+
+def get_selectable_changes(
+    selectables_from,
+    selectables_target,
+    enums_from,
+    enums_target,
+    add_dependents_for_modified=True,
+):
+    tables_from = od((k, v) for k, v in selectables_from.items() if v.is_table)
+    tables_target = od((k, v) for k, v in selectables_target.items() if v.is_table)
+
+    other_from = od((k, v) for k, v in selectables_from.items() if not v.is_table)
+    other_target = od((k, v) for k, v in selectables_target.items() if not v.is_table)
+
+    added_tables, removed_tables, modified_tables, unmodified_tables = differences(
+        tables_from, tables_target
+    )
+    added_other, removed_other, modified_other, unmodified_other = differences(
+        other_from, other_target
+    )
+
+    changed_all = {}
+    changed_all.update(modified_tables)
+    changed_all.update(modified_other)
+    modified_all = dict(changed_all)
+    changed_all.update(removed_tables)
+    changed_all.update(removed_other)
+
+    replaceable = set()
+    not_replaceable = set()
+
+    if add_dependents_for_modified:
+        for k, m in changed_all.items():
+            old = selectables_from[k]
+
+            if k in modified_all and m.can_replace(old):
+                if not m.is_table:
+                    replaceable.add(k)
+                continue
+
+            for d in m.dependents_all:
+                if d in unmodified_other:
+                    dd = unmodified_other.pop(d)
+                    modified_other[d] = dd
+                not_replaceable.add(d)
+        modified_other = od(sorted(modified_other.items()))
+
+    replaceable -= not_replaceable
+    statements = Statements()
+
+    def functions(d):
+        return {k: v for k, v in d.items() if v.relationtype == "f"}
+
+    statements += statements_from_differences(
+        added_other,
+        removed_other,
+        modified_other,
+        replaceable=replaceable,
+        drops_only=True,
+        dependency_ordering=True,
+        old=selectables_from,
+    )
+
+    statements += get_table_changes(
+        tables_from, tables_target, enums_from, enums_target
+    )
+
+    if any([functions(added_other), functions(modified_other)]):
+        statements += ["set check_function_bodies = off;"]
+
+    statements += statements_from_differences(
+        added_other,
+        removed_other,
+        modified_other,
+        replaceable=replaceable,
+        creations_only=True,
+        dependency_ordering=True,
+        old=selectables_from,
+    )
     return statements
 
 
 class Changes(object):
-
     def __init__(self, i_from, i_target):
         self.i_from = i_from
         self.i_target = i_target
 
     def __getattr__(self, name):
-        if name == 'schema':
-            return partial(
-                get_schema_changes,
-                self.i_from.tables,
-                self.i_target.tables,
-                self.i_from.enums,
-                self.i_target.enums,
-            )
-
-        elif name == 'non_pk_constraints':
+        if name == "non_pk_constraints":
             a = self.i_from.constraints.items()
             b = self.i_target.constraints.items()
             a_od = od((k, v) for k, v in a if v.constraint_type != PK)
             b_od = od((k, v) for k, v in b if v.constraint_type != PK)
             return partial(statements_for_changes, a_od, b_od)
 
-        elif name == 'pk_constraints':
+        elif name == "pk_constraints":
             a = self.i_from.constraints.items()
             b = self.i_target.constraints.items()
             a_od = od((k, v) for k, v in a if v.constraint_type == PK)
             b_od = od((k, v) for k, v in b if v.constraint_type == PK)
             return partial(statements_for_changes, a_od, b_od)
 
-        elif name == 'views_and_functions':
-            av = self.i_from.views.items()
-            bv = self.i_target.views.items()
-            af = self.i_from.functions.items()
-            bf = self.i_target.functions.items()
-            avf = list(av) + list(af)
-            bvf = list(bv) + list(bf)
-            avf = od(sorted(avf))
-            bvf = od(sorted(bvf))
+        elif name == "selectables":
             return partial(
-                statements_for_changes, avf, bvf, add_dependents_for_modified=True
+                get_selectable_changes,
+                od(sorted(self.i_from.selectables.items())),
+                od(sorted(self.i_target.selectables.items())),
+                self.i_from.enums,
+                self.i_target.enums,
             )
 
         elif name in THINGS:
